/**
 *
 * $Id: ansi.hpp,v 1.6 2004/12/06 07:31:05 mviara Exp $
 * $Author: mviara $
 * $Name:  $
 *
 * $Log: ansi.hpp,v $
 * Revision 1.6  2004/12/06 07:31:05  mviara
 * Improved performance.
 *
 * Revision 1.5  2004/11/30 20:05:28  mviara
 * Version 2.02
 *
 * Revision 1.4  2004/11/10 18:40:17  mviara
 * Version 2.01
 *
 * Revision 1.3  2004/07/12 15:59:35  mviara
 * Removed same rem, added supporto for Visual C and Watcom C.
 *
 * Revision 1.2  2004/07/10 13:59:58  mviara
 * Add Supporto for connection time.
 *
 * Revision 1.1.1.1  2004/07/10 12:33:21  mviara
 * Migrated from RCS
 *
 * Revision 1.1  2001/08/08 16:14:42  VM
 * Initial revision
 *
 * 
 */
#include <windows.h>

#include <dos.h>
#include <assert.H>
#include <stdio.h>
#include <string.h>
#include <conio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <fcntl.h>
#include <io.h>
#include <time.h>
#include <ctype.h>
#include <assert.h>


#define	copyright	"ANSIX364/32 Vers. 2.03 - Created by Mario Viara - mario@viara.cn"
void AnsiPrintf(char * , ...);
void AnsiPutchar(char c);
void AnsiError(char *, ...);



//
// Struttura per definire un carattere nel buffer video
//
typedef struct
{
	unsigned char chr;		// carattere
	unsigned char  color;		// attributo
} SCHAR;




//
// Struttura per gli adatattori video
//

class CAdapter
{
protected:
	short		numRows;	// numero di righe
	short		numCols;	// numero di colonne
	unsigned	curColor;	// colore corrente
public:
	CAdapter();
	virtual char * Name();		// nome del driver
	virtual void SetColor(unsigned col);
	virtual void Locate(int pos)=0;
	virtual void PrintBuffer(int pos,char * buffer,int num);
	virtual void Print(int pos,char  ch,int num)=0;
	virtual int Kbhit();		// torna 1 se c'e' un tasto
	virtual unsigned Getch();	// legge un tasto
	virtual void HideCursor();	// nasconde il cursore
	virtual void ShowCursor();	// visualizza il cursore
	int Rows() {return numRows;}	// torna il numero di righe
	int Cols() {return numCols;}	// torna il numero di colonne
	int Size() {return numCols*numRows;};
};

class CAdapBIOS : public CAdapter
{
private:
	virtual char * Name();
	virtual void Locate(int pos);
	virtual void Print(int pos,char  ch,int num);
public:
	CAdapBIOS();

};

class CAdaGRAPH : public CAdapter
{
private:
	unsigned oldCurColor;
	virtual char * Name();
	virtual void Locate(int pos);
	virtual void Print(int pos,char  ch,int num);
public:
	CAdaGRAPH();
};

class CAdapterNT : public CAdapter
{
	CONSOLE_CURSOR_INFO	cursorInfo;
	
	unsigned Getkey(int kbdtype,int padnum);
private:
	int	kbdType;
	int	padnum;
	unsigned oldCurColor;
	HANDLE hStdOut;
	HANDLE hStdIn;
	DWORD  oldConsoleOutMode;
	DWORD  oldConsoleInMode;
	unsigned key;
	virtual char * Name();
	virtual void SetColor(unsigned col);
	virtual void Locate(int pos);
	virtual void Print(int pos,char  ch,int num);
	virtual void PrintBuffer(int pos,char *buffer,int num);
	virtual int Kbhit();
	virtual unsigned Getch();
	virtual void HideCursor();
	virtual void ShowCursor();
	void	SetAttribute();

public:
	CAdapterNT(int keyboard,int p);
	~CAdapterNT();
};

class CDevice
{
	protected:
	char	proxyName[256];
	int	flagProxy;
	
public:
	CDevice();
	void	SetProxy(char *);
	virtual int  IsConnected(int code);
	virtual void Deactivate(int code);	// deattiva la session
	virtual int Connect(int code,char *id);		// connette la sessione
	virtual void Disconnect(int code);	// disconnette la sessione
	virtual int Read(char *buffer,int maxSize)=0;
	virtual int Write(char *buffer,int size)=0;
	virtual char * Name() ;			// torna il nome del driver
};

class CDeviceECHO : public CDevice
{
	char buffer[256];
	int size;
public:
	CDeviceECHO();
	virtual int Read(char *buffer,int maxSize);
	virtual int Write(char *buffer,int size);
	virtual char * Name() ;
};

class CDeviceINT14 : public CDevice
{
	int port;
public:
	CDeviceINT14() {port=0;};
	virtual int Read(char *buffer,int maxSize);
	virtual int Write(char *buffer,int size);
	virtual char * Name() ;
};



	
class CAnsi
{
public:
	class CScreen
	{
		public:
		enum
		{
			KEYBUF = 127,		// Deve essere 2^n - 1
			KEYLEN = 16,		// lunghezza tasti funzione
			FCNKEY= 71
		};
	
		int	numRow,numCol;
		SCHAR * buffer;		
		int	session;
		void	(CAnsi::*DisplayChar)(unsigned char);
		int	nopar;
		int	numpar;		/* numero di parametri x comando	*/
		int	par[8];		/* parametri comando ansi		*/
		int wrapped;
		int *	parptr;		/* pointer parametri			*/
		int	vscr_nopar;	/* flag nessun parametro		*/
		unsigned start;		/* start address			*/
		int	ps;		/* pointer scritura buffer tastiera	*/
		int	pl;		/* pointer lettura buffer tastiera	*/
		unsigned char	keybuf[KEYBUF+1];
		unsigned char  keys[FCNKEY][KEYLEN]; // tasti funzione		*/
		unsigned char lenght[FCNKEY];	/* lunghezze tasti funzione	*/
		int	reverse;	/* flag reverse				*/
		int	blink;
		int	hi;
		int	gvt100;
		int	autoWrap;
		unsigned bnormal,fnormal;
		unsigned  breverse,freverse;
		unsigned bcolor,fcolor;
		unsigned bgraphics,fgraphics; // 15.01.98
		long	lptCounter;
		SCHAR   *lptVideoBuffer;
		unsigned lptCursor;
		char	lptTmp[256];
		FILE	*fdlpt;
		int	printchar;	/* caratteri da stampare ma non ancora	*/
						/* stampati perche' uguali alla fine	*/
						/* stampa.				*/
		int	bscroll;	/* riga iniziale scroll			*/
		int	escroll;	/* riga finale scroll			*/

		int	cursorOn;
		int	curPos;
		int	oldCurPos;
		unsigned	curColor;
		unsigned	orMask;
		char flagScrollOff;	// scroll disabilitato
		void SetColor(unsigned col) {curColor = col;};
		void Locate(int pos) {curPos = pos;};
		void Print(char ch,int num);
		void Print(int pos,char  ch,int num){Locate(pos);Print(ch,num);};
		CScreen(int row,int col);
		~CScreen();
		void Default();
	};

	enum
	{
		MAXSCREENS = 10,
	        KEYSCROLLOFF = 0xff00,
		KEYSCROLLON  = 0xff01,			     
	};

	enum EMode
	{
		Ansix364,
		Vt100,
		Linux
	};
	
private:
	friend void AnsiPrintf(char *,...);
	friend void AnsiError(char *,...);
	friend void AnsiPutchar(char);
	
	static CAnsi * defaultAnsi;
	SCHAR	* tmpBuffer;
	EMode	terminalMode;
	int	doubleBuffer;
	int	selectPrinter;
	long	deviceRead;
	long	deviceWrite;
	CScreen * curScr;
	CScreen * screens[MAXSCREENS+1];
	time_t	screenTime[MAXSCREENS+1];
	unsigned startConnect[MAXSCREENS+1];
	unsigned endConnect[MAXSCREENS+1];
	CAdapter * adapter;
	CDevice  * device;
	int	flagSwitch;						// Flag switch screen attivo
	int	flagExit ;
	int	fdCapture;
	int	fdFromFile;
	int	numRows;
	int	numCols;
	int	screenSize;
	char	printerName[64];
	
public:
	void HideCursor();
	void ShowCursor();
	void Macro();
	void Capture();
	CAnsi(CAdapter * a,CDevice *d,EMode  ,int=0,int=0);
	~CAnsi();
	void SetColor(unsigned color)
	{
		curScr->SetColor(color);
		adapter->SetColor(color);
	};
			
	void ResetScreen();
	int  ChangeScreen();
	void AddScreen(int s);
	int  SwitchScreen(int s);
	void ActivateScreen(int s);
	void ConnectScreen(int s);
	void DeactivateScreen(int s);
	void DeleteScreen();
	void Main();
        void Putchar(char);
	void Outstr(char *buffer,int len);
	void Outstr(char * buffer) {Outstr(buffer,strlen(buffer));};
	void CalcAttribute();
	void PrinterOutput(unsigned char c);
	void DisplaySpecial(unsigned char c);
	void SelectPrinter(char*);
	void StartPrinter();
	void StopPrinter();
	void DisplayPrinter(unsigned char c);
	void DisplayNormal(unsigned char c);
	void DisplayEscape(unsigned char c);
	void DisplayCsi(unsigned char c);
	void DisplayAdditional(unsigned char c);
	void DisplayFcnkey(unsigned char c);
	void DisplayPrgkey(unsigned char c);
	void DisplayVt100Mode(unsigned char c);
	void ProcessAttribute(int n);
	void CsiAttribute();
	void OutChar(unsigned char c);
	void Bell() {putch(7);};
	void BackSpace();
	void Tab();
	void BackTab(int ) {}; //bug
	void LineFeed();
	void CarriageReturn();
	void ScrollUp(int n);
	void ScrollDn(int n);
	void VideoCopy(int source,int dest,int count);
	void VideoGet(int pos,int n,SCHAR * buffer);
	void VideoClear(int r,int c,int nr,int nc);
	void VideoPut(int pos,int n,SCHAR * buffer);
	void VideoSet(int pos,char c,int n);
	void Locate();
	void StartEscape() {curScr->DisplayChar = DisplayEscape;};
	void StartCsi() ;
	int GetPar(char c);
	void Inschar(int col,int no,int eol)	;
	void Delchar(int col,int no,int eol)	;
	void KeybInsert(unsigned k);
	void KeybInsert(char *);
	void KeybInput();
	int  KeybStatus();
	int  KeybRead();
	int  Read(char *,int);
	int  Write(char * ,int);
	void Info();
	void PrintScreen();
	void SetPrinterName(char *);
	void SetSwitch(int flag);
	char * TermId();
};

class CDeviceTelnet : public CDevice
{
	typedef struct
	{
		SOCKET	socket;
		int	echoTX;
		int	echoRX;
		int	isValid;
		char	bufferIn[512];
		int	readBufferIn;
		int	numBufferIn;
	} STelnetSession;
	
	char * termType;
	char hostName[64];
	STelnetSession	telnetSession[CAnsi::MAXSCREENS];
	STelnetSession *curSession;
	int session;
	enum
	{
		TELOPT_ECHO = 1,
		TELOPT_TTYPE = 24,
		DO = 253,
		WILL = 251,
		WONT = 252,
		DONT = 254,
		IAC = 255,
		SB = 250,
		SE = 240,
	};
	int WriteRaw(unsigned char *buffer,int size);	
	void Send3(int n1,int n2,int n3);
	void Dont(int c) {Send3(IAC,DONT,c);};
	void Will(int c) {Send3(IAC,WILL,c);};
	void Wont(int c) {Send3(IAC,WONT,c);};
	void Do(int c)   {Send3(IAC,DO,c);};
	int  RecvChar();
	void SBReceived();
	void WILLReceived();
	void WONTReceived();
	void DOReceived();
	int  IACReceived();
	void DONTReceived();
	int CheckSocket();
public:
	CDeviceTelnet(char * hostname);
	virtual ~CDeviceTelnet();
	virtual int	IsConnected(int code);
	virtual int		Connect(int code,char *id);
	virtual void Deactivate(int code);
	virtual void Disconnect(int code);
	virtual int Read(char *buffer,int maxSize);
	virtual int Write(char *buffer,int size);
	virtual char * Name() ;
};


#define	APPCODE	0x19980120
#define APPNAME "ansi32"
	
